#include <sys/keycodes.h>

#include <assert.h>
#include <display.h>
#include <err.h>
#include <errno.h>
#include <pixel.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <timespec.h>

#define PLAYFIELD_SIDE 100
#define TILE_SIDE 10
#define TILES (PLAYFIELD_SIDE / TILE_SIDE)
enum palette {
	BG = 0,
	TEXT,
	GRASS_TOP,
	GRASS_BROWN,
	ERROR,
};
static enum palette playfield[PLAYFIELD_SIDE * PLAYFIELD_SIDE];

static char tilemap[] = "          "
                        "          "
                        "          "
                        "          "
                        "gg        "
                        "          "
                        "          "
                        "     g    "
                        "    g g   "
                        "gggggggggg";

static uint32_t main_window = 0;
static uint32_t window_width = 600;
static uint32_t window_height = 500;

static bool game_running = true;

static void on_disconnect(void *context) {
	(void) context;
	game_running = false;
}

static bool left_held, right_held, jump_held;

static void on_keyboard(void *context, uint32_t window, uint32_t codepoint) {
	(void) context;
	if (window != main_window)
		return;
	int key = KBKEY_DECODE(codepoint);
	switch (key) {
		// TODO: rebinding
		case KBKEY_LEFT: left_held = true; break;
		case -KBKEY_LEFT: left_held = false; break;
		case KBKEY_RIGHT: right_held = true; break;
		case -KBKEY_RIGHT: right_held = false; break;
		case KBKEY_SPACE: jump_held = true; break;
		case -KBKEY_SPACE: jump_held = false; break;
	}
}

static void on_quit(void *context, uint32_t window) {
	(void) context;
	if (window != main_window)
		return;
	game_running = false;
}

static void on_resize(void *context, uint32_t window, uint32_t width, uint32_t height) {
	(void) context;
	if (window != main_window)
		return;
	window_width = width;
	window_height = height;
}

static double timespec2double(struct timespec ts) {
	return ts.tv_sec + ts.tv_nsec / 1000.0 / 1000.0 / 1000.0;
}

char sample_tilemap(double x, double y) {
	size_t tx = x / TILE_SIDE;
	size_t ty = y / TILE_SIDE;
	assert(tx < TILES);
	assert(ty < TILES);
	return tilemap[ty * TILES + tx];
}

static double player_x, player_y;
static double player_dx, player_dy;
static void update(struct timespec now, struct timespec dt_timespec) {
	(void) now;
	double dt = timespec2double(dt_timespec);

	player_x += player_dx * dt;
	player_y += player_dy * dt;

	if (left_held) player_dx = -30;
	else if (right_held) player_dx = 30;
	else player_dx = 0;

	// Collision against walls of the map
	if (player_x < 0) {
		player_x = 0;
		player_dx = 0;
	} else if (player_x > PLAYFIELD_SIDE - TILE_SIDE) {
		player_x = PLAYFIELD_SIDE - TILE_SIDE;
		player_dx = 0;
	}
	if (player_y < 0) {
		player_y = 0;
		player_dy = 0;
	} else if (player_y > PLAYFIELD_SIDE - TILE_SIDE) {
		player_y = PLAYFIELD_SIDE - TILE_SIDE;
		player_dy = 0;
	}

	// Collision against tiles
	char topleft = sample_tilemap(player_x, player_y);
	char topright = sample_tilemap(player_x + TILE_SIDE, player_y);
	char bottomleft = sample_tilemap(player_x, player_y + TILE_SIDE);
	char bottomright = sample_tilemap(player_x + TILE_SIDE, player_y + TILE_SIDE);

	bool on_ground = false;
	if (bottomleft != ' ' || bottomright != ' ')
		on_ground = true;
	if (topleft != ' ' || topright != ' ' || bottomleft != ' ' || bottomright != ' ')
		player_dx = 0;

	if (on_ground)
		player_dy = 0;
	else
		player_dy = 30;

}

static void draw_grass_tile(size_t x, size_t y) {
	assert(x + TILE_SIDE <= PLAYFIELD_SIDE);
	assert(y + TILE_SIDE <= PLAYFIELD_SIDE);
	char grass_tile[] = "xxxxxxxxxx"
	                    "xx xx x x "
	                    " x  x   x "
	                    "          "
	                    "          "
	                    "          "
	                    "          "
	                    "          "
	                    "          "
	                    "          ";
	assert(strlen(grass_tile) == TILE_SIDE * TILE_SIDE);
	for (size_t ty = 0; ty < TILE_SIDE; ty++) {
		for (size_t tx = 0; tx < TILE_SIDE; tx++) {
			enum palette pixel;
			switch (grass_tile[ty * TILE_SIDE + tx]) {
				case 'x': pixel = GRASS_TOP; break;
				case ' ': pixel = GRASS_BROWN; break;
				default: pixel = ERROR; break;
			}
			playfield[(y + ty) * PLAYFIELD_SIDE + x + tx] = pixel;
		}
	}
}

static void draw_tiles(void) {
	assert(strlen(tilemap) == TILES * TILES);
	for (size_t ty = 0; ty < TILES; ty++) {
		for (size_t tx = 0; tx < TILES; tx++) {
			switch(tilemap[ty * TILES + tx]) {
				case 'g': draw_grass_tile(tx * TILE_SIDE, ty * TILE_SIDE); break;
			}
		}
	}
}

static void draw(void) {
	memset(playfield, 0, sizeof(playfield));
	draw_tiles();
	draw_grass_tile(player_x, player_y);
}

int main(int argc, char *argv[]) {
	struct display_connection *connection = display_connect_default();
	if (!connection && errno == ECONNREFUSED)
		display_spawn(argc, argv);
	if (!connection)
		err(1, "Could not connect to display server");

	display_create_window(connection, main_window);
	display_resize_window(connection, main_window, window_width, window_height);
	display_show_window(connection, main_window);

	struct display_event_handlers handlers = {0};
	handlers.disconnect_handler = on_disconnect;
	handlers.keyboard_handler = on_keyboard;
	handlers.quit_handler = on_quit;
	handlers.resize_handler = on_resize;

	uint32_t palette_colours[] = {
		[BG] = make_color(128, 128, 255),
		[TEXT] = make_color(255, 255, 255),
		[GRASS_TOP] = make_color(0, 255, 0),
		[GRASS_BROWN] = make_color(200, 128, 0),
		[ERROR] = make_color(255, 0, 255),
	};

	long fps_cap = 60;
	struct timespec min_dt = timespec_make(0, 1000L * 1000L * 1000L / fps_cap);
	struct timespec last_frame;
	clock_gettime(CLOCK_MONOTONIC, &last_frame);

	int fps = 0;
	while (game_running) {
		struct timespec now;
		clock_gettime(CLOCK_MONOTONIC, &now);
		struct timespec dt = timespec_sub(now, last_frame);
		if (timespec_lt(dt, min_dt)) {
			struct timespec sleep = timespec_sub(min_dt, dt);
			nanosleep(&sleep, NULL);
			continue;
		}

		if (last_frame.tv_sec != now.tv_sec) {
			char *title = NULL;
			asprintf(&title, "%i fps", fps);
			display_title_window(connection, main_window, title);
			free(title);
			fps = 0;
		} else
			fps++;

		update(now, dt);
		draw();

		size_t pixels;
		if (__builtin_mul_overflow(window_width, window_height, &pixels))
			errx(1, "Window size too big");
		uint32_t *framebuffer = calloc(pixels, sizeof(uint32_t));

		uint32_t shorter_side = window_height < window_width ? window_height : window_width;
		size_t y_offset = (window_height - shorter_side) / 2;
		size_t x_offset = (window_width - shorter_side) / 2;
		for (size_t y = 0; y < window_height; y++) {
			for (size_t x = 0; x < window_width; x++) {
				if (y < y_offset || x < x_offset ||
				    y >= y_offset + shorter_side || x >= x_offset + shorter_side)
					framebuffer[y * window_width + x] = make_color_a(0, 0, 0, 200);
				else {
					size_t pf_x = (x - x_offset) * PLAYFIELD_SIDE / shorter_side;
					size_t pf_y = (y - y_offset) * PLAYFIELD_SIDE / shorter_side;
					uint32_t colour = palette_colours[playfield[pf_y * PLAYFIELD_SIDE + pf_x]];
					framebuffer[y * window_width + x] = colour;
				}
			}
		}
		display_render_window(connection, main_window, 0, 0, window_width, window_height, framebuffer);
		free(framebuffer);

		while (!display_poll_event(connection, &handlers));
		
		last_frame = now;
	}

	display_disconnect(connection);

	return 0;
}
